import argparse
import hashlib
import logging
import os

import pyftpdlib.authorizers
import pyftpdlib.handlers
import pyftpdlib.servers

LOG_LEVELS = {
    'critical': logging.CRITICAL,
    'error': logging.ERROR,
    'warning': logging.WARNING,
    'info': logging.INFO,
    'debug': logging.DEBUG,
}


class SHA256Authorizer(pyftpdlib.authorizers.DummyAuthorizer):

    def validate_authentication(self, username, password, handler):
        # pyftpdlib/authorizers.py", line 110, in add_user
        #   dic = {'pwd': str(password),
        # so we can not compare against .digest()
        password_hash = hashlib.sha256(password.encode()).hexdigest()
        if self.user_table[username]['pwd'] != password_hash.lower():
            raise pyftpdlib.authorizers.AuthenticationFailed()


def serve(root_dir_path, username, password_sha256_hexdigest,
          control_port, passive_port, passive_address, umask, log_level):
    logging.basicConfig(level=log_level)
    if umask:
        os.umask(umask)
        logging.info('umask=0o%o', umask)
    assert os.path.isdir(root_dir_path), root_dir_path
    authorizer = SHA256Authorizer()
    authorizer.add_user(
        username,
        password_sha256_hexdigest.lower(),
        homedir=root_dir_path,
        # https://pyftpdlib.readthedocs.io/en/latest/api.html#pyftpdlib.authorizers.DummyAuthorizer.add_user
        # e: change dir
        # m: mkdir
        # w: write
        perm='emw',
        msg_login='renshi ni hen gaoxing',
        msg_quit='zaijian',
    )
    handler = pyftpdlib.handlers.FTPHandler
    handler.authorizer = authorizer
    handler.banner = 'ni hao'
    handler.passive_ports = (passive_port,)
    handler.masquerade_address = passive_address
    server = pyftpdlib.servers.FTPServer((None, control_port), handler)
    # apparently requires +1 for unknown reasons
    server.max_cons = 1 + 1
    server.serve_forever()


class EnvDefaultArgparser(argparse.ArgumentParser):

    def add_argument(self, *args, envvar=None, **kwargs):
        if envvar:
            envvar_value = os.environ.get(envvar, None)
            if envvar_value:
                kwargs['required'] = False
                kwargs['default'] = envvar_value
        super().add_argument(*args, **kwargs)


def _init_argparser():
    argparser = EnvDefaultArgparser()
    argparser.add_argument(
        '--root', '--root-dir',
        metavar='path',
        dest='root_dir_path',
        default=os.getcwd(),
        help='default: current working directory',
    )
    argparser.add_argument(
        '--user', '--username',
        metavar='username',
        dest='username',
        required=True,
        envvar='FTP_USERNAME',
        help='default: env var $FTP_USERNAME',
    )
    argparser.add_argument(
        '--pwd-hash', '--password-hash',
        metavar='sha256_hexdigest',
        dest='password_sha256_hexdigest',
        required=True,
        envvar='FTP_PASSWORD_SHA256',
        help='default: env var $FTP_PASSWORD_SHA256',
    )
    argparser.add_argument(
        '--ctrl-port', '--control-port',
        metavar='port',
        type=int,
        dest='control_port',
        envvar='FTP_CONTROL_PORT',
        default=2121,
        help='default: env var $FTP_CONTROL_PORT or 2121',
    )
    argparser.add_argument(
        '--pasv-port', '--passive-port',
        metavar='port',
        type=int,
        dest='passive_port',
        envvar='FTP_PASSIVE_PORT',
        default=62121,
        help='port for passive (PASV) & extended passive (EPSV) mode;'
        + ' default: env var $FTP_PASSIVE_PORT or 62121',
    )
    argparser.add_argument(
        '--pasv-addr', '--passive-address', '--masquerade-address',
        metavar='ip_address',
        dest='passive_address',
        envvar='FTP_PASSIVE_ADDRESS',
        default=None,
        help='address returned to client (227)'
        + ' after opening port for passive mode (PASV)'
        + ' default: socket\'s own address',
    )
    argparser.add_argument(
        '--umask',
        type=int,
        envvar='UMASK',
        default=' default: env var $UMASK',
    )
    argparser.add_argument(
        '--log-level',
        metavar='level_name',
        dest='log_level_name',
        default='info',
        choices=LOG_LEVELS.keys(),
        help='default: %(default)s',
    )
    return argparser


def main():
    args = _init_argparser().parse_args()
    args.log_level = LOG_LEVELS[args.log_level_name]
    del args.log_level_name
    serve(**vars(args))



from setuptools import setup

setup(
    name='pyftpd-sink',
    use_scm_version=True,
    description='write-only ftp server',
    author='Fabian Peter Hammerle',
    author_email='fabian@hammerle.me',
    url='https://git.hammerle.me/fphammerle/pyftpd-sink',
    license='MIT',
    keywords=[
        'ftp',
        'server',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Internet :: File Transfer Protocol (FTP)',
        'Programming Language :: Python',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ],
    packages=[],
    entry_points={
        'console_scripts': [
            'pyftpd-sink = pyftpd_sink:main',
        ],
    },
    install_requires=[
        'pyftpdlib>=1.5.4',
    ],
    setup_requires=[
        'setuptools_scm',
    ],
    tests_require=[
        'pytest',
    ],
)



# pylint: disable=missing-docstring

import unittest.mock

import pytest

import pyftpd_sink


@pytest.mark.parametrize(('env', 'argv', 'expected_umask'), [
    ({}, [], None),
    ({}, ['--umask', str(0o0027)], 0o0027),
    ({}, ['--umask', str(0o0177)], 0o0177),
    ({}, ['--umask', '23'], 0o0027),
    ({}, ['--umask', '127'], 0o0177),
    ({'UMASK': str(0o0022)}, [], 0o0022),
    ({'UMASK': str(0o0022)}, ['--umask', str(0o0027)], 0o0027),
])
def test_main_umask(env, argv, expected_umask):
    with unittest.mock.patch('pyftpd_sink.serve') as serve_mock:
        with unittest.mock.patch('sys.argv', ['', '--user', 'u', '--pwd-hash', 'p'] + argv):
            with unittest.mock.patch('os.environ', env):
                pyftpd_sink.main()
    serve_mock.assert_called_once()
    serve_kwargs = serve_mock.call_args[1]
    assert expected_umask == serve_kwargs['umask']
